import io
import threading
import os
from io import DEFAULT_BUFFER_SIZE
from time import sleep

ModeR = 0
ModeW = 1


class BufferObject(object):
    """Extend of io.BytesIO
    """
    
    def __init__(self, data=None, size=-1):
        """Constructor

        Args:
        data: initial data
        size: buffer size
        """

        self._size = size
        self._buf = io.BytesIO()
        self._len = 0
        self._w_pos = 0
        self._r_pos = 0
        self._lock = threading.Lock()
        self._mode = ModeW
        if data:
            self.write(data)

    def close(self):
        self._buf.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return False

    @property
    def length(self):
        return self._len

    @property
    def rest(self):
        """readable size"""
        return self._len - self._r_pos

    @property
    def lack(self):
        """lack of required write size"""
        return self._size - self._w_pos

    def set_r_mode(self):
        self._mode = ModeR

    def set_w_mode(self):
        self._mode = ModeW

    def tell(self):
        if self._mode == ModeR:
            return self._r_pos
        else:
            return self._w_pos

    def seek(self, pos, whence=os.SEEK_SET):
        if self._mode == ModeR:
            if whence == os.SEEK_SET:
                pass
            elif whence == os.SEEK_CUR:
                pos += self._r_pos
            elif whence == os.SEEK_END:
                pos += self._size
            self._r_pos = sorted([0, pos, self._size])[1]
            return self._r_pos
        elif self._mode == ModeW:
            if whence == os.SEEK_SET:
                pass
            elif whence == os.SEEK_CUR:
                pos += self._w_pos
            elif whence == os.SEEK_END:
                pos += self._w_pos
            if pos < 0:
                pos = 0
            if pos > self._len:
                pos = self._len
            self._w_pos = sorted([0, pos, self._len])[1]
            return self._w_pos

    def read(self, size=-1):
        with self._lock:
            self._buf.seek(self._r_pos)
            ret = self._buf.read(size)
            self._r_pos = self._buf.tell()
            return ret

    def write(self, data):
        with self._lock:
            self._buf.seek(self._w_pos)
            self._buf.write(data)
            self._w_pos = self._buf.tell()
            self._buf.seek(0, os.SEEK_END)
            self._len = self._buf.tell()

    def clear(self):
        self._buf.close()
        del self._buf

        self._buf = io.BytesIO()
        self._len = 0
        self._w_pos = 0
        self._r_pos = 0


class BufferObjectR(BufferObject):
    """Read-Purpose BufferObject
    Can be written only once at first
    """
    
    def __init__(self, data):
        super(BufferObjectR, self).__init__()
        super(BufferObjectR, self).write(data=data)

    @property
    def pos(self):
        return self._r_pos

    def write(self, data):
        raise RuntimeError("can not add to buffer")

    def set_w_mode(self):
        raise RuntimeError("can not set write mode")


class BufferObjectW(BufferObject):
    """Write-Purpose BufferObject
    Can be read only once at last
    """
    
    def __init__(self, size):
        super(BufferObjectW, self).__init__(size=size)

    @property
    def pos(self):
        return self._w_pos

    def read(self, size=-1):
        if size > 0:
            raise RuntimeError("can not read partially")
        return super(BufferObjectW, self).read(size)

    def set_r_mode(self):
        raise RuntimeError("can not set read mode")


class ChunkedBuffer(BufferObject):
    """Chunked-BufferObject
    Constituted by individual closed possible plurality of buffers
    """
    def __init__(
            self,
            size=-1,
            buffer_size=DEFAULT_BUFFER_SIZE,
            read_at_first=False,
            retry_limit=-1,
            delay=0.1
    ):
        """Constructor
        
        Args:
        size (int): size to be written
        buffer_size (int): buffer(chunk) size
        read_at_first (bool): 
        if True, when can read from chunk, wait to write
        retry_limit (int): retry limit
        if less then 0, wait everytime
        delay (float): retry delay
        """
        self._buffer_size = buffer_size
        super(ChunkedBuffer, self).__init__()
        self._size = size
        self._read_at_first = read_at_first
        self._retry_limit = retry_limit
        self._r_pos_1 = 0
        self._r_pos_2 = 0
        self._w_pos_1 = 0
        self._w_pos_2 = 0
        self._buffers = []
        self._add_chunk()
        self._kill_flag = False
        self._delay = delay
        self._closed_pos = 0

        if size < 0:
            raise ValueError("Invalid size")

    @property
    def lack(self):
        if self._w_pos_1 > self._r_pos_1 and self._w_pos > self._r_pos + self._buffer_size:
            return 0
        else:
            return min(self._buffer_size - self._w_pos_2, self._size - self._w_pos)

    @property
    def _r_pos(self):
        return self._r_pos_1 * self._buffer_size + self._r_pos_2

    @_r_pos.setter
    def _r_pos(self, pos):
        self._r_pos_1 = pos // self._buffer_size
        self._r_pos_2 = pos % self._buffer_size

    @property
    def _w_pos(self):
        return self._w_pos_1 * self._buffer_size + self._w_pos_2

    @_w_pos.setter
    def _w_pos(self, pos):
        self._w_pos_1 = pos // self._buffer_size
        self._w_pos_2 = pos % self._buffer_size

    def _add_chunk(self):
        size = min(self._buffer_size, self._size - self._w_pos)
        self._buffers.append(BufferObject(size=size))

    def close(self):
        self._kill_flag = True
        with self._lock:
            for buf in self._buffers:
                buf.close()
                del buf

    def close_chunk(self):
        if self._r_pos_1 > self._closed_pos:
            with self._lock:
                for buf in self._buffers[self._closed_pos:self._r_pos_1]:
                    buf.close()
                    del buf
                self._closed_pos = self._r_pos_1

    def read(self, size=-1):
        try:
            if size < 0:
                size = self._size
            size = min(size, self._size - self._r_pos)
            w_buf = BufferObjectW(size=size)
            retries = 0
            flg_continue = False
            while w_buf.lack > 0:
                # Force Quit
                if self._kill_flag:
                    break

                if retries > self._retry_limit >= 0:
                    raise Exception("Retry count exceed the limit in read()")

                r_buf = self._buffers[self._r_pos_1]
                with self._lock:
                    r_buf._r_pos = self._r_pos_2
                    if r_buf.rest == 0:
                        if self._r_pos_1 < self._w_pos_1:
                            self._r_pos_1 += 1
                            self._r_pos_2 = 0
                            r_buf = self._buffers[self._r_pos_1]
                        else:
                            retries += 1
                            flg_continue = True

                if flg_continue:
                    sleep(self._delay)
                    flg_continue = False
                else:
                    with self._lock:
                        r_size = min(r_buf.rest, w_buf.lack)
                        data = r_buf.read(r_size)
                        w_buf.write(data)
                        self._r_pos_2 += len(data)
                        retries = 0

            data = w_buf.read()
            w_buf.close()
            return data
        except Exception:
            self._kill_flag = True
            raise

    def write(self, data):
        try:
            r_buf = BufferObjectR(data=data)
            retries = 0
            flg_continue = False
            while r_buf.rest > 0:
                # Force Quit
                if self._kill_flag:
                    break

                if retries > self._retry_limit >= 0:
                    raise Exception("Retry count exceed the limit in write()")

                if self._read_at_first and self.lack == 0:
                    retries += 1
                    flg_continue = True

                if flg_continue:
                    sleep(self._delay)
                    flg_continue = False
                else:
                    w_buf = self._buffers[self._w_pos_1]
                    with self._lock:
                        w_buf._w_pos = self._w_pos_2

                        r_size = min(r_buf.rest, w_buf.lack)
                        data = r_buf.read(r_size)
                        w_buf.write(data)
                        retries = 0

                        if w_buf.lack == 0:
                            self._add_chunk()
                            self._w_pos_1 += 1
                            self._w_pos_2 = 0
                        else:
                            self._w_pos_2 = w_buf._w_pos

            r_buf.close()
            self._len = self._w_pos
        except Exception:
            self._kill_flag = True
            raise

    def clear(self):
        raise NotImplementedError
