# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""OneDrive File Access module.

Define classes for OneDrive file access.
"""

# common module
from __future__ import unicode_literals
from fileaccess import (FileObj, Access)
import os
import stat
import arrow
from buffer_object import BufferObject, BufferObjectW, ChunkedBuffer
from fileaccess_helper import *

# feature module
import re
from arrow.parser import ParserError as ArrowParseError
from buffer_object import BufferObjectR
from buffer_object2 import ChunkedBuffer
import threading
import sys
from hashlib import md5
import requests
import datetime
import urllib
import json
import httplib
import time
import urlparse
import random

# common constant
FILE_SEPARATOR = u"/"
FILE_MODE = 0
DIR_MODE = stat.S_IFDIR
ERR_MSG_RENAME = "Invalid new name"
ERR_MSG_INVALID_FILE_LENGTH = "Invalid file length"
ERR_MSG_INVALID_OPEN_MODE = "Invalid open mode"
OPEN_MODE_READ = "r"
OPEN_MODE_WRITE = "w"

# feature constant
SESSION_PATH_PERM = (stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP |
                        stat.S_IWGRP | stat.S_IROTH | stat.S_IWOTH)
PROTOCOL_LIST = ("http", "https")
ACCESS_TOKEN_NAME = "access_token"
REFRESH_TOKEN_NAME = "refresh_token"
UNAUTHORIZED_RETRY_COUNT = 1
SERVER_ERR_RETRY_COUNT = 3
RETRY_SLEEP_TIME = 30
THROTTLED_ERR_RETRY_COUNT = 5
EXP_BACKOFF_BASE = 16


class NeedToRetry(BaseException): pass

class OneDriveDownloader(object):
    def __init__(self, path, access, length, buffer_size):
        self.path = path
        self.access = access
        self.length = length
        self.buffer_size = buffer_size
        
        target_item = self.access._get_file(self.path)
        self.downloadUrl = target_item['@microsoft.graph.downloadUrl']
        self.pos = 0

    def close(self):
        pass

    def cancel(self):
        pass

    def next(self):
        range_from = self.pos
        range_to = min(self.length, self.pos + self.buffer_size) - 1
        range = range_to - range_from
        count = 0
        while True:
            if range < 0:
                raise StopIteration()
            data = self.read_range(range_from, range_to)
            if len(data) == range + 1:
                self.pos += range + 1
                return data
            if count > SERVER_ERR_RETRY_COUNT:
                raise Exception("unmatch download size: %d(%d-%d)" % (range, range_from, range_to))

    def read_range(self, range_from, range_to):
        headers = {
            'Range': "bytes=%d-%d" % (range_from, range_to),
        }
        data = self.access._get(
            self.downloadUrl, 
            headers=headers,
            use_api=False)
        return data

class OneDriveUploader(object):
    def __init__(self, path, access, length, buffer_size):
        self.path = path
        self.access = access
        self.length = length
        self.buffer_size = buffer_size
        self.buffer = None
        self.exception = None
        self.is_alive = True
        self.is_complete = False

    def close(self): pass

    def upload(self): pass

class OneDriveEmptyUploader(OneDriveUploader):
    def __init__(self, path, access, length, buffer_size):
        super(OneDriveEmptyUploader, self).__init__(path, access, length, buffer_size)
        self.upload()

    def close(self):
        pass

    def upload(self):
        target_item = self.access._get_file(self.path)
        if target_item is None:
            parent_path = self.access._get_directory_name(
                os.path.dirname(self.path))
            parent_item = self.access._get_file(parent_path, be_raise=True)
            target_name = os.path.basename(self.path)
            self.access._put('/me/drive/items/{0}:/{1}:/content'.format(
                parent_item['id'], target_name), data='')
        else:
            r = self.access._put('/me/drive/items/{0}/content'.format(
                target_item['id']), data='')

    def write(self, data):
        pass


class OneDriveStandardUploader(OneDriveUploader):
    def __init__(self, path, access, length, buffer_size):
        super(OneDriveStandardUploader, self).__init__(path, access, length, buffer_size)
        self.buffer = ChunkedBuffer(size=self.length, buffer_size=self.buffer_size)
        self.thread = threading.Thread(target=self.upload)
        self.thread.start()

    def close(self):
        if self.is_alive:
            self.thread.join()
        self.buffer.close()
        if self.exception:
            raise self.exception[1], None, self.exception[2]

    def upload(self):
        try:
            self.buffer.set_r_mode()
            target_item = self.access._get_file(self.path)
            if target_item is None:
                parent_path = self.access._get_directory_name(
                    os.path.dirname(self.path))
                parent_item = self.access._get_file(parent_path, be_raise=True)
                target_name = os.path.basename(self.path)
                self.access._put('/me/drive/items/{0}:/{1}:/content'.format(
                    parent_item['id'], target_name), data=self.buffer)
            else:
                self.access._put('/me/drive/items/{0}/content'.format(
                    target_item['id']), data=self.buffer)
            self.is_complete = True
        except Exception as e:
            self.exception = sys.exc_info()
            self.buffer.close()
            self.is_alive = False

    def write(self, data):
        if self.is_alive:
            self.buffer.write(data)


class OneDriveMultiPartUploader(OneDriveUploader):
    def __init__(self, path, access, length, buffer_size):
        super(OneDriveMultiPartUploader, self).__init__(path, access, length, buffer_size)
        self.upload_size = 0
        self.buffer = BufferObjectW(size=min(buffer_size, self.length))

        target_item = self.access._post(
            '/me/drive/root{0}/createUploadSession'.format(
                self.access._get_api_file_path(self.path)))
        self.upload_url = target_item['uploadUrl']

    def close(self):
        if self.is_complete:
            self.complete()
        else:
            self.cancel()
        self.buffer.close()

    def upload(self):
        data = self.buffer.read()
        range = 'bytes {0}-{1}/{2}'.format(
            self.upload_size, self.upload_size + self.buffer.length - 1, self.length)
        headers = {
            'Content-Length': str(self.length),
            'Content-Range' : range,
        }
        self.access._put(self.upload_url, use_api=False, data=data, headers=headers)
        
        self.upload_size += self.buffer.length
        if self.upload_size == self.length:
            self.is_complete = True

    def write(self, data):
        r_buf = BufferObjectR(data=data)
        while r_buf.rest > 0:
            data = r_buf.read(self.buffer.lack)
            self.buffer.write(data)
            if self.buffer.lack == 0:
                self.upload()
                self.buffer.close()
                self.buffer = BufferObjectW(size=min(
                    self.buffer_size, self.length - self.upload_size))
        r_buf.close()

    def complete(self):
        self.buffer.close()

    def cancel(self):
        self.access._delete(self.upload_url, use_api=False)
        self.buffer.close()

class OneDriveFileObj(FileObj):
    """Class for OneDrive file."""

    def __init__(self, access, path, mode, length, chunk_size):
        """Constructor."""

        super(OneDriveFileObj, self).__init__()
        self._access = access
        self._path = path
        self._mode = (filter((lambda m: m in mode.lower()), (OPEN_MODE_READ, OPEN_MODE_WRITE)) or [None])[0]
        self._length = length
        self._chunk_size = chunk_size
        if self._chunk_size is None:
            self._chunk_size = 4 * 1024 * 1024
        self._r_pos = 0
        self._w_pos = 0
        self._downloader = None
        self._uploader = None
        self._buffer = None

        if self._mode == OPEN_MODE_READ:
            self._length = self._access.stat_file(self._path)[stat.ST_SIZE]
            self._download()
        elif self._mode == OPEN_MODE_WRITE:
            if self._length < 0:
                raise ValueError(ERR_MSG_INVALID_FILE_LENGTH)
            self._upload()
        else:
            raise ValueError(ERR_MSG_INVALID_OPEN_MODE)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return False

    def close(self):
        """Close file."""

        try:
            if self._mode == OPEN_MODE_WRITE:
                if self._uploader:
                    self._uploader.close()
            else:
                if self._downloader:
                    self._downloader.close()
        finally:
            if self._buffer:
                self._buffer.close()

    def read(self, size=-1):
        """Return a "size" bytes data from current position in file.
        If the size argument is negative or omitted, then return all data.
        """

        if size < 0:
            read_size = self._length
        else:
            read_size = min(size, self._length - self._r_pos)

        ret_buf = BufferObjectW(size=read_size)
        while ret_buf.lack > 0:
            data = self._buffer.read(ret_buf.lack)
            if not len(data):
                break

            ret_buf.write(data)
            if self._buffer.rest == 0:
                self._buffer.close()
                self._download()

        self._r_pos += ret_buf.length
        ret = ret_buf.read()
        ret_buf.close()
        
        return ret

    def _download(self):
        """Download to buffer"""

        if not self._downloader:
            target_item = self._access._get_file(self._path, be_raise=True)
            if target_item is not None:
                self._length = self._access._stat(target_item)[stat.ST_SIZE]
                self._buffer = BufferObject(size=self._chunk_size)
                self._downloader = OneDriveDownloader(
                    path=self._path,
                    access=self._access,
                    length=self._length,
                    buffer_size=self._chunk_size,
                )

        self._buffer.set_w_mode()
        self._buffer.clear()
        while self._buffer.lack > 0:
            try:
                data = self._downloader.next()
                self._buffer.write(data)
            except StopIteration:
                break
        self._buffer.set_r_mode()

    def write(self, data):
        """Write a "data" data to current position in file."""
        self._uploader.write(data)

    def _upload(self):
        if not self._uploader:
            if self._length == 0:
                self._uploader = OneDriveEmptyUploader(
                    path=self._path,
                    access=self._access,
                    length=self._length,
                    buffer_size=self._chunk_size,
                )
            elif self._length >= 4 * 1024 * 1024:
                self._uploader = OneDriveMultiPartUploader(
                    path=self._path,
                    access=self._access,
                    length=self._length,
                    buffer_size=self._chunk_size,
                )
            else:
                self._uploader = OneDriveStandardUploader(
                    path=self._path,
                    access=self._access,
                    length=self._length,
                    buffer_size=self._chunk_size,
                )


class OneDriveAccess(Access):
    """class accessing to OneDrive file."""

    def __init__(self, client_id, client_secret, base_url, authorize_base_url,
            token_base_url, redirect_url, scope, session_path=None, 
            proxy=None, proxy_port=None, proxy_user=None, proxy_pass=None):
        """Constructor.

        Args:
            client_id: client id
            client_secret: client secret
            base_url: Graph API endpoint
            authorize_base_url: OAuth2 authorize endpoint
            token_base_url: OAuth2 token endpoint 
            redirect_url: OAuth2 redirect url
            session_path: path for session info(token/code)
            proxy: Address/hostname for a proxy server
                If None, bypass proxies
            proxy_port: The port to use when connecting over a proxy
            proxy_user: The username to connect with on the proxy
            proxy_pass: The password to use when connection over a proxy
        """

        super(OneDriveAccess, self).__init__()

        self.session = requests.Session()
        self.session.proxies.update(self._get_proxies_dict(PROTOCOL_LIST,
            proxy, proxy_port, proxy_user, proxy_pass))
        
        self._client_id = client_id
        self._client_secret = client_secret
        self._base_url = base_url
        self._authorize_base_url = authorize_base_url
        self._token_base_url = token_base_url
        self._redirect_url = redirect_url
        self._scope = scope
        self._session_path = session_path

        self._access_token_info = {}
        
    def _get_proxies_dict(self, protocol_list, proxy=None,
             proxy_port=None, proxy_user=None, proxy_pass=None):
        """generate dict about proxy infomation for Requests
        Returns:
            dict:
        """
        _proxies = {}
        if proxy:
            parse_result = urlparse.urlparse(proxy)
            if parse_result.hostname:
                proxy = parse_result.hostname
            if proxy_port is None:
                proxy_port = parse_result.port
            if proxy_user is None:
                proxy_user = parse_result.username
            if proxy_pass is None:
                proxy_pass = parse_result.password
            if proxy_user and proxy_pass:
                proxy = '%s:%s@%s' % (proxy_user, proxy_pass, proxy)
            if proxy_port:
                proxy = '%s:%s' % (proxy, proxy_port)
            proxies = {}
            for protocol in protocol_list:
                _proxies[protocol] = \
                            urlparse.urlunparse((protocol, proxy, "", "", "", ""))
        return _proxies

    @property
    def max_num_transfer_threads(self):
        """Return maximum thread number for file transfer."""

        return 5

    @property
    def auth_url(self):
        param = {
            'client_id'    : self._client_id,
            'response_type': 'code',
            'redirect_uri' : self._redirect_url,
            'scope'        : self._scope,
            'state'        : 12345,
        }
        return self._authorize_base_url + '?' + urllib.urlencode(param)

    def authenticate(self, params=None):
        if params is None:
            params = {}
        code = params.get('code')
        if code is not None:
            self._access_token_info = self._get_access_token_info(code)
            self._save_access_token_info()
        else:
            self._load_access_token_info()
        
    def _get_access_token_info(self, code):
        """"Return a access token data for the content API of Graph API server."""
        data = {
            'grant_type'   : 'authorization_code',
            'code'         : code,
            'scope'        : self._scope,
            'client_id'    : self._client_id,
            'client_secret': self._client_secret,
            'redirect_uri' : self._redirect_url,
        }
        r = self.session.post(self._token_base_url, data = {
                'grant_type'   : 'authorization_code',
                'code'         : code,
                'scope'        : self._scope,
                'client_id'    : self._client_id,
                'client_secret': self._client_secret,
                'redirect_uri' : self._redirect_url,
            })
        return self._parse(r)

    def _refresh_access_token_info(self):
        """Refresh a access token and return it."""
        r = self.session.post(self._token_base_url, data={
                'grant_type'    : 'refresh_token',
                'refresh_token' : self._access_token_info[REFRESH_TOKEN_NAME],
                'scope'         : self._scope,
                'client_id'     : self._client_id,
                'client_secret' : self._client_secret,
                'redirect_uri'  : self._redirect_url,
            })
        self._access_token_info = self._parse(r)
        self._save_access_token_info()

    def _save_access_token_info(self):
        """Save current access token info to session file."""
        with open(self._session_path, "w") as session_file:
            session_file.write(json.dumps(self._access_token_info))
        os.chmod(self._session_path, SESSION_PATH_PERM)

    def _load_access_token_info(self):
        """Load access token info from session file."""
        with open(self._session_path) as session_file:
            self._access_token_info = json.loads(session_file.read())

    def _get(self, url, **kwargs):
        return self._request('GET', url, **kwargs)

    def _post(self, url, **kwargs):
        return self._request('POST', url, **kwargs)

    def _put(self, url, **kwargs):
        return self._request('PUT', url, **kwargs)

    def _patch(self, url, **kwargs):
        return self._request('PATCH', url, **kwargs)

    def _delete(self, url, **kwargs):
        return self._request('DELETE', url, **kwargs)

    def _request(self, method, url, headers=None, **kwargs):
        _headers = {'Accept': 'application/json'}
        use_api = kwargs.pop('use_api', True)
        if use_api:
           url = self._base_url + url
           _headers.update({'Authorization': 'Bearer ' + self._access_token_info[ACCESS_TOKEN_NAME]})
        if headers:
            _headers.update(headers)
        if 'files' not in kwargs:
            _headers['Content-Type'] = 'application/json'
        stream = kwargs.get('stream', None)
        chunk_size = kwargs.pop('chunk_size', None)
        
        count_svr_err = 0
        count_unauth = 0
        count_thr = 0
        while True:
            try:
                return self._parse(self.session.request(method, url, headers=_headers, **kwargs), stream, chunk_size)
            except requests.exceptions.HTTPError as e:
                try:
                    if (e.response.status_code == httplib.INTERNAL_SERVER_ERROR or
                        e.response.status_code == httplib.BAD_GATEWAY or
                        e.response.status_code == httplib.SERVICE_UNAVAILABLE or
                        e.response.status_code == httplib.GATEWAY_TIMEOUT):
                        if count_svr_err < SERVER_ERR_RETRY_COUNT:
                            count_svr_err += 1
                            time.sleep(RETRY_SLEEP_TIME)
                            raise NeedToRetry()
                    elif e.response.status_code == httplib.UNAUTHORIZED:
                        if count_unauth < UNAUTHORIZED_RETRY_COUNT:
                            count_unauth += 1
                            self._refresh_access_token_info()
                            _headers.update({'Authorization': 
                                'Bearer ' + self._access_token_info[ACCESS_TOKEN_NAME]})
                            raise NeedToRetry()
                    elif e.response.status_code == 429:
                        if count_thr < THROTTLED_ERR_RETRY_COUNT:
                            count_thr += 1
                            _sleep_time = e.response.headers.get('Retry-After')
                            if _sleep_time is None:
                                _base_time = EXP_BACKOFF_BASE * (2 ** count_thr)
                                _jitter = random.randint(-_base_time / 10, _base_time / 10)
                                _sleep_time = _base_time+ _jitter
                            else:
                                _sleep_time = int(_sleep_time)
                            time.sleep(_sleep_time)
                            raise NeedToRetry()
                    e.args = (e.args if e.args else tuple()) + \
                        ('content=' + e.response.content,)
                    raise
                except NeedToRetry:
                    data = kwargs.get("data")
                    if isinstance(data, ChunkedBuffer):
                        data.set_r_mode()
                        data.seek(0)
                    continue

    def _parse(self, response, stream=None, chunk_size=None):
        response.raise_for_status()
        content_type = response.headers.get('Content-Type')
        if stream:
            r = response.iter_content(chunk_size=chunk_size)
        elif content_type is not None and 'application/json' in content_type:
            r = response.json()
        else:
            r = response.content
        return r

    @staticmethod
    def _get_api_file_path(file_path):
        """Get URL string for path to file in Graph API

        Args:
             file_path (unicode): path
        Returns:
            string:
        """
        _file_path = OneDriveAccess._get_file_name(file_path)
        if _file_path == '':
            return ''
        _url_path = urllib.quote(_file_path.encode('utf-8'))
        _api_path = ':{0}{1}:'.format(FILE_SEPARATOR, _url_path)
        return _api_path

    @staticmethod
    def _get_file_name(path):
        """Get path of "file" for OneDrive

        Args:
            path (str or unicode): path strings
        Returns:
            unicode:
        """
        if isinstance(path, str):
            path = path.decode('utf-8')
        return path.strip(FILE_SEPARATOR)

    @staticmethod
    def _get_directory_name(path):
        """Get path of "directory" for OneDrive

        Args:
            path (str or unicode): path strings
        Returns:
            unicode:
        """
        return OneDriveAccess._get_file_name(path) + FILE_SEPARATOR

    def _get_file(self, file_path, be_raise=False):
        """Get file object

        Args:
            file_path (unicode): path
            be_raise (bool): If the item is not found,
                an exception is raised if is True,
                or return None if is False
        Returns:
            DriveItem if found, None if not Found
        """
        try:
            item = self._get('/me/drive/root{0}'.format(
                self._get_api_file_path(file_path)))
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 404:
                item = None
            else:
                raise

        if item is None:
            if be_raise:
                raise LookupError(file_path)
            else:
                return None
        return item

    def _get_directory(self, file_path, be_raise=False):
        """Get directory object

        Args:
            file_path (unicode): path
            be_raise (bool): If the item is not found,
                an exception is raised if is True,
                or return None if is False
        Returns:
            DriveItem if found, None if not Found
        """
        
        return self._get_file(file_path, be_raise)

    @staticmethod
    def _parse_timestamp(target_str):
        if target_str is not None:
            try:
                return arrow.get(target_str).timestamp
            except ArrowParseError:
                try:
                    return arrow.get(target_str, "ddd, DD MMM YYYY HH:mm:ss ZZZ").timestamp
                except ArrowParseError:
                    return arrow.get(target_str, "ddd, DD MMM YYYY HH:mm:ss").replace(tzinfo="+09:00").timestamp
        return None

    @staticmethod
    def _stat(target_item):
        """Return a stat information dictionary for "target_item".

        Args:
            target_item : target DriveItem
        Returns:
            dict: stat of item

            {
                stat.ST_MODE: mode,
                stat.ST_MTIME: mtime,
                stat.ST_CTIME: ctime,
                stat.ST_SIZE: size,
            }

                {
                    stat.ST_MODE: 0 if target is file, stat.S_IFDIR if target is directory,

                    stat.ST_MTIME: modified time for target,

                    stat.ST_CTIME: changed time for target, but S3 is not support,

                    stat.ST_SIZE: file size if target is file, 0 if target is directory,

                }
        """

        if target_item.get('folder') is None:
            _mode = FILE_MODE
        else:
            _mode = DIR_MODE

        _mtime = OneDriveAccess._parse_timestamp(
            target_item['lastModifiedDateTime'])
        _ctime = OneDriveAccess._parse_timestamp(
            target_item['createdDateTime'])
        _size = None
        if _mode == FILE_MODE:
            _size = target_item.get('size')

        return {
            stat.ST_MODE: _mode,
            stat.ST_MTIME: _mtime,
            stat.ST_CTIME: _ctime,
            stat.ST_SIZE: _size,
        }

    def make_directory(self, dir_path):
        """Create "dir_path" directory."""

        dir_path = self._get_file_name(dir_path)
        
        if dir_path == "" or dir_path == FILE_SEPARATOR:
            # if blank, do nothing
            return

        target_id = self._get_directory(dir_path, be_raise=False)
        if target_id is not None:
            # if exists, do nothing
            return

        parent_path = self._get_file_name(
            os.path.dirname(dir_path))
        parent_item = self._get_directory(parent_path)
        if parent_item is None:
            self.make_directory(parent_path)
            parent_item = self._get_directory(parent_path)

        new_dir_name = os.path.basename(dir_path)
        data = {
            'name': new_dir_name, 
            'folder': { },
            '@microsoft.graph.conflictBehavior': "rename"
        }
        self._post('/me/drive/items/{0}/children'.format(
            parent_item['id']), data=json.dumps(data))

    def delete_directory(self, dir_path):
        """Delete "dir_path" directory."""
        
        self.delete_file(dir_path)

    def stat_directory(self, dir_path):
        """Return a stat information dictionary for "file_path"
           which keys below:
            stat.ST_MODE, stat.ST_MTIME, stat.CTIME, stat.ST_SIZE
        """

        return self.stat_file(dir_path)

    def rename_directory(self, dir_path, new_name):
        """Rename a name of "dir_path" to "new_name"."""

        if FILE_SEPARATOR in new_name:
            raise Exception(ERR_MSG_RENAME, {"new_name": new_name})
        
        self.rename_file(dir_path, new_name)

    def move_directory(self, dir_path, new_dir_path):
        """Move "dir_path" to "new_path"."""

        self.move_file(dir_path, new_dir_path)

    def get_contents(self, dir_path, with_stat=False):
        """Return a contents dictionary in "dir_path".
           A stat information must be included, if "with_stat" is True, 
        """

        contents = {}
        target_item = self._get_directory(dir_path)
        data = self._get('/me/drive/items/{0}/children'.format(target_item['id']))
        while True:
            children = data['value']
            for item in children:
                name = item['name']
                stat = {}
                if with_stat:
                    stat = self._stat(item)
                contents[name] = stat
            nextLink = data.get("@odata.nextLink")
            if nextLink is None:
                break
            data = self._get(nextLink, use_api=False)
        return contents

    def get_updated_contents(self, dir_path):
        """Return a updated contents dictionary in "dir_path"."""

        raise NotImplementedError

    def open(self, file_path, mode="r", length=-1, chunk_size=None):
        """Return a class instance of the "FileObj" or sub class
           corresponding to "file_path".
        """

        return OneDriveFileObj(
            access=self,
            path=self._get_file_name(file_path),
            mode=mode,
            length=length,
            chunk_size=chunk_size,
        )

    def delete_file(self, file_path):
        """Delete "file_path" file."""

        target_item = self._get_file(file_path, be_raise=True)
        if target_item is not None:
            self._delete('/me/drive/items/{0}'.format(target_item['id']))

    def stat_file(self, file_path):
        """Return a stat information dictionary for "file_path"
           which keys below:
            stat.ST_MODE, stat.ST_MTIME, stat.CTIME, stat.ST_SIZE
        """

        target_item = self._get_file(file_path, be_raise=True)
        return self._stat(target_item)

    def rename_file(self, file_path, new_name):
        """Rename a name of "file_path" to "new_name"."""

        if FILE_SEPARATOR in new_name:
            raise Exception("Invalid new_name", {"new_name": new_name})

        target_item = self._get_file(file_path, be_raise=True)
        data = {
            'name': new_name,
        }
        self._patch('/me/drive/items/{0}'.format(target_item['id']), data=json.dumps(data))

    def move_file(self, file_path, new_dir_path):
        """Move "file_path" to "new_path"."""
        
        target_item = self._get_file(file_path, be_raise=True)
        target_dir  = self._get_directory(new_dir_path, be_raise=True)

        data = {
            'parentReference': {
              'id': target_dir['id'],
            },
        }
        self._patch('/me/drive/items/{0}'.format(target_item['id']), data=json.dumps(data))

    def test(self):
        pass
